#include "md3tomdl.h"


bool loadskin(std::ifstream &pcxfile, skin &modelskin)
{
	pcx_header head;

	pcxfile.read((char*)&head, sizeof(pcx_header));

	std::cout<<sizeof(pcx_header) << " bytes read" << std::endl;
	/*
	cout << (int)head.bpp << endl;
	cout << (int)head.bytesperline << endl;
	cout << (int)head.encoding << endl;
	cout << (int)head.hdpi << endl;
	cout << (int)head.vdpi << endl;
	cout << (int)head.hscreensize << endl;
	cout << (int)head.vscreensize << endl;
	cout << (int)head.version << endl;
	cout << (int)head.manufacturer << endl;
	cout << (int)head.nplanes << endl;
	cout << (int)head.paletteinfo << endl;
	cout << (int)head.reserved << endl;
	cout << (int)head.xmax << endl;
	cout << (int)head.xmin << endl;
	cout << (int)head.ymax << endl;
	cout << (int)head.ymin << endl;
	*/

	if (head.bpp != 8)
	{
		std::cout << "not an 8bit pcx file" << std::endl;
		return false;
	} 

	modelskin.width = (1 + head.xmax - head.xmin);
	modelskin.height = (1 + head.ymax - head.ymin);
	modelskin.skindata = new char[ (modelskin.width * modelskin.height) * 2 ];

	int32_t i = 0;
	
	while(i < (modelskin.width * modelskin.height) )
	{
		int8_t b[1];
		int32_t btop;
		int8_t c[1];
		pcxfile.read((char*)b,1);
		btop = *b & 0xC0;
		if(btop == 0xC0  ) //if both of the top two bits are set
		{
			pcxfile.read((char*)c,1); //read the next int8_t and repeat it the number of times in the 6 lower bits

			int32_t count;
			count = (*b & 0x3F);
//			std::cout << count << std::endl;
			for(int32_t j = 0; j < count ; ++j)
			{
				modelskin.skindata[i] = *c;
				++i;
			}
		}
		else // just write that character
		{
				modelskin.skindata[i] = *b;
				++i;
		}
	}




	return true;
}


uint8_t findnormal (q3_vertex v)
{
	float theta = (float)(v.norm >> 8) * M_PI / 255;
	float phi = (float)(v.norm & 255) * M_PI / 255;

	vec3_t normal;
	normal[0] = cos( theta) * sin (phi);
	normal[1] = sin( theta) * sin (phi);
	normal[2] = cos (phi);

	uint8_t best;
	float bestproduct;
	best = 0;
	bestproduct = anorms_table[0][0] * normal[0] + anorms_table[0][1] * normal[1] + anorms_table[0][2] * normal[2];
	for(int32_t i = 0; i < 162; ++i)
	{
		float currentproduct = anorms_table[i][0] * normal[0] + anorms_table[i][1] * normal[1] + anorms_table[i][2] * normal[2];
		if(currentproduct > bestproduct)
		{
			bestproduct = currentproduct;
			best = (unsigned char) i;
		}
	}
	return best;
}


void adjust_bounds(q3_frame &frame, q3_vertex &vert)
{
	float coords[3];

	coords[0] = (float) vert.x * MD3_XYZ_SCALE + frame.org[0];
	coords[1] = (float) vert.y * MD3_XYZ_SCALE + frame.org[1];
	coords[2] = (float) vert.z * MD3_XYZ_SCALE + frame.org[2];

	if(frame.mins[0] > frame.maxs[0]) //these being backwards tells us no vertices have been processed for this frame
	{
		frame.mins[0] = frame.maxs[0] = coords[0];
		frame.mins[1] = frame.maxs[1] = coords[1];
		frame.mins[2] = frame.maxs[2] = coords[2];
		return;
	}

	if(frame.mins[0] > coords[0])
		frame.mins[0] = coords[0];
	if(frame.mins[1] > coords[1])
		frame.mins[1] = coords[1];
	if(frame.mins[2] > coords[2])
		frame.mins[2] = coords[2];

	if(frame.maxs[0] < coords[0])
		frame.maxs[0] = coords[0];
	if(frame.maxs[1] < coords[1])
		frame.maxs[1] = coords[1];
	if(frame.maxs[2] < coords[2])
		frame.maxs[2] = coords[2];
/*
	cout << frame.maxs[0] << " ";
	cout << frame.maxs[1] << " ";
	cout << frame.maxs[2] << " to ";
	cout << frame.mins[0] << " ";
	cout << frame.mins[1] << " ";
	cout << frame.mins[2] << " " << endl;*/
}


void parse_surface(std::ifstream &q3input, q3_surface& surf, q3_model &q3mdl)
{
	//std::cout << " Parsing surface " << surf.header.name << ":" << std::endl;
	
	// triangles
	q3input.seekg(q3mdl.header.ofs_surfaces + surf.header.ofs_triangles);
	for(int32_t i = 0; i < surf.header.num_triangles; i++)
	{
		q3_triangle tri;
		q3input.read((char*)&tri, sizeof(q3_triangle));
		surf.triangle.push_back(tri);
	}

	// vertices - is there a need to match up these with the triangle indices explicitly?
	q3input.seekg(q3mdl.header.ofs_surfaces + surf.header.ofs_vertices);
	surf.frame.resize(surf.header.num_frames);

	for(int32_t i = 0; i < surf.header.num_frames; i++)
	{
		for(int32_t j = 0; j < surf.header.num_verts; j++)
		{
			q3_vertex vert;
			q3input.read((char*)&vert, sizeof(q3_vertex));
			surf.frame[i].vertex.push_back(vert); //add it to the last element of surf.frame
			adjust_bounds( q3mdl.frames[i],surf.frame[i].vertex[j]);
		}
	}

	//check this stuff is working
	//std::cout << surf.frame.size() << " frames in this surface" << std::endl;

	// uv map
	q3input.seekg(q3mdl.header.ofs_surfaces + surf.header.ofs_st);
	for(int32_t i = 0; i < surf.header.num_verts; i++)
	{
		q3_texcoord uv;
		q3input.read((char*)&uv, sizeof(q3_texcoord));
		surf.texcoord.push_back(uv);
	}
}


int32_t parse_q3model(std::ifstream &q3input, q3_model &q3mdl)
{
	//Check for correct file format
	int8_t ident[5];
	q3input.read((char*)ident,4);
	ident[4] = 0;
	if (strcmp((char*)ident,"IDP3"))
	{
		std::cout << "Input is not a valid md3 model, ident value:" << ident << std::endl;
		return 0;
	}
	
	//read the version, although we don't need it
	int32_t version;
	q3input.read((char*)&version,4);
	std::cout << "model is version " << version << std::endl;

	//skip the path
	q3input.ignore(64); 

	//load flags
	q3input.read((char*)&q3mdl.header.flags, 4);
	//std::cout << "model flags: " << q3mdl.header.flags << std::endl;
	//load number of frames
	q3input.read((char*)&q3mdl.header.framecount , 4);
	std::cout << "model frames: " << q3mdl.header.framecount << std::endl;
	//skip number of tags (discarded in q1 format)
	q3input.ignore(4); 
	//load number of surfaces
	q3input.read((char*)&q3mdl.header.surfacecount , 4);
	std::cout << "model surfaces: " << q3mdl.header.surfacecount << std::endl;
	//discard number of skins
	q3input.ignore(4); 
	//load offsets
	q3input.read((char*)&q3mdl.header.ofs_frames , 4);
	//std::cout << "ofs_frames: " << q3mdl.header.ofs_frames << std::endl;
	q3input.read((char*)&q3mdl.header.ofs_surfaces , 4);
	//std::cout << "ofs_surfaces: " << q3mdl.header.ofs_surfaces << std::endl;

	// load the frames
	q3input.seekg(q3mdl.header.ofs_frames);
	//std::cout << " frames loaded:" << std::endl;

	for( int32_t i = 0; i < q3mdl.header.framecount; i++)
	{
		q3_frame currentframe;
		q3input.read((char*)&currentframe, sizeof(q3_frame));
		currentframe.mins[0] = 1;
		currentframe.maxs[0] = -1;

		q3mdl.frames.push_back(currentframe);
	//	std::cout << currentframe.name << std::endl;
	}

	// load the surface headers
	q3input.seekg(q3mdl.header.ofs_surfaces);
	//std::cout << " surfaces loaded:" << std::endl;

	for( int32_t i = 0; i < q3mdl.header.surfacecount; i++)
	{
		q3_surfaceheader currentsurfheader;
		q3_surface currentsurf;
		q3input.read((char*)&currentsurfheader, sizeof(q3_surfaceheader));

		currentsurf.header = currentsurfheader;
		q3mdl.surfaces.push_back(currentsurf);
		//std::cout << currentsurf.header.name << std::endl;
		q3input.seekg(q3mdl.header.ofs_surfaces + currentsurf.header.ofs_end); // skip to start of next header

	}
	//now actually load the contents of each surface

	std::vector<q3_surface>::iterator surfiter;
	for (surfiter = q3mdl.surfaces.begin(); surfiter != q3mdl.surfaces.end(); surfiter++)
	{
		parse_surface(q3input, *surfiter, q3mdl);
	}
	/*
	for each (q3_surface surf in q3mdl.surfaces)
		cout << surf.frame.size() << " frames in list" << endl;
	*/


	return 1;

}

void findbounds(float* mins, float* maxs, float& rad, q3_model& q3mdl)
{
	mins[0] = q3mdl.frames[0].mins[0];
	mins[1] = q3mdl.frames[0].mins[1];
	mins[2] = q3mdl.frames[0].mins[2];

	maxs[0] = q3mdl.frames[0].maxs[0];
	maxs[1] = q3mdl.frames[0].maxs[1];
	maxs[2] = q3mdl.frames[0].maxs[2];

	rad = 0;
	for each (q3_frame f in q3mdl.frames)
	{
		if(f.mins[0] < mins[0])
			mins[0] = f.mins[0];
		if(f.mins[1] < mins[1])
			mins[1] = f.mins[1];
		if(f.mins[2] < mins[2])
			mins[2] = f.mins[2];

		if(f.maxs[0] > maxs[0])
			maxs[0] = f.maxs[0];
		if(f.maxs[1] > maxs[1])
			maxs[1] = f.maxs[1];
		if(f.maxs[2] > maxs[2])
			maxs[2] = f.maxs[2];

		if(f.radius > rad)
			rad = f.radius;
	}

}

void writedummyskin(std::ofstream& mdlfile) //export a dummy skin
{
	int32_t skintype = 0;
	int8_t check[17] = "aaaaaaaaZZZZZZZZ";//make a checkerboard pattern
	mdlfile.write( (char*)&skintype ,4);
	for(int32_t i = 0; i < (256*16) ; i++)
		mdlfile.write((char*)check,16);
}

static int32_t q1_skinheight = 256;
static int32_t q1_skinwidth = 256;

void export_to_mdl(std::ofstream& mdlfile, q3_model& q3mdl, skin& q1skin)
{
	float mins[3];
	float maxs[3];
	q1_header head;
	findbounds(mins, maxs, head.boundingradius, q3mdl);
	/*
	std::cout << mins[0] << " ";
	std::cout << mins[1] << " ";
	std::cout << mins[2] << " to ";
	std::cout << maxs[0] << " ";
	std::cout << maxs[1] << " ";
	std::cout << maxs[2] << std::endl;
	*/

	head.scale[0] = (maxs[0] - mins[0]) / 255;
	head.scale[1] = (maxs[1] - mins[1]) / 255;
	head.scale[2] = (maxs[2] - mins[2]) / 255;
	/*
	std::cout << "scale: " << head.scale[0] << " ";
	std::cout << head.scale[1] << " ";
	std::cout << head.scale[2] << std::endl;
	*/

	head.translate[0] = mins[0];
	head.translate[1] = mins[1];
	head.translate[2] = mins[2]; //head.radius is added to the header directly by findbounds

	head.ident = 1330660425; //move to constructor?
	head.version = 6;

	head.eyeposition[0] = head.eyeposition[1] = 0;
	head.eyeposition[2] = q3mdl.frames[0].org[2]; //might as well grab it from first frame

	head.num_skins = 1;
	if(!q1skin.height)
	{
		head.skinheight = q1_skinheight; //magic!
		head.skinwidth = q1_skinwidth;
	}
	else
	{
		head.skinheight = q1skin.height; //magic!
		head.skinwidth = q1skin.width;
	}
	int32_t vertcount = 0;
	int32_t tricount = 0;

	std::vector<q3_surface>::iterator surfiter;
	for (surfiter = q3mdl.surfaces.begin(); surfiter != q3mdl.surfaces.end(); surfiter++)
	{
		vertcount += (*surfiter).header.num_verts;
		tricount += (*surfiter).header.num_triangles;
	}
	head.num_verts = vertcount;
	head.num_tris = tricount;
	head.num_frames = (int)q3mdl.frames.size();

	head.synctype = 1; //almost always preferable
//	head.flags = q3mdl.header.flags;
	head.flags = 0;
	head.size = 1; //arbitrary

	mdlfile.write((char*)&head, sizeof(head));
	//skin
	if(!q1skin.height)
		writedummyskin(mdlfile);
	else
	{
		std::cout << "writing pcx skin, width: " << q1skin.width <<", height: " << q1skin.height << std::endl;
		if(!q1skin.skindata)
			std::cout << "skindata corrupt" << std::endl;
		int32_t skintype = 0;
		mdlfile.write( (char*)&skintype ,4);
		for(int32_t i = 0; i < (q1skin.height * q1skin.width) ; ++i)
			mdlfile.write(q1skin.skindata+i, 1);
	}


	//uv coords

	//int32_t uvtracker = 0;

	for (surfiter = q3mdl.surfaces.begin(); surfiter != q3mdl.surfaces.end(); surfiter++)
	{
		int32_t type = 0;
		for each(q3_texcoord uv in (*surfiter).texcoord)
		{
			int32_t u = (int)(uv.u *head.skinwidth);
			int32_t v = (int)(uv.v *head.skinheight);


			u = std::max<int32_t>(0,std::min(u,head.skinwidth)); //clamping to range
			v = std::max<int32_t>(0,std::min(v,head.skinheight));
			
			mdlfile.write((char*)&type, 4);
			mdlfile.write((char*)&u, 4);
			mdlfile.write((char*)&v, 4);
	//		++uvtracker;
		}
	}
	//std::cout << uvtracker << " uvs written of " << head.num_verts << endl;

	//int32_t tritracker = 0;
	int32_t triindexbump = 0; //increase the index of triangle vertices in later objects
	//triangles
	for (surfiter = q3mdl.surfaces.begin(); surfiter != q3mdl.surfaces.end(); surfiter++)
	{
		int32_t type = 0;
		for each(q3_triangle tri in (*surfiter).triangle)
		{
			q3_triangle temptri = tri;
			temptri.index[0] += triindexbump;
			temptri.index[1] += triindexbump;
			temptri.index[2] += triindexbump;
			mdlfile.write((char*)&type, 4);
			mdlfile.write((char*)&temptri.index[0], 4);
			mdlfile.write((char*)&temptri.index[1], 4);//note that this is going to fail to give sensible values
			mdlfile.write((char*)&temptri.index[2], 4);//when model has multiple surfaces!
//			cout << temptri.index[0] << " " << temptri.index[1] << " " << temptri.index[2] << endl;


	//		++tritracker;
		}
		triindexbump += (*surfiter).header.num_verts;

	}
	//cout << tritracker << " tris written of " << head.num_tris << endl;


	//frames
	
	int32_t type = 0;
	uint8_t maxcomp[4] = {255, 255, 255, 0}; //compressed bounding box format
	uint8_t mincomp[4] = {0, 0, 0, 0}; //this is actually a cheat which may be fixed later
	int32_t i = 0;
	uint8_t coords[4]; //mostly to 'intialize' the vertex normal to 0, not yet computed correctly
//	coords[3] = 0;
	for each(q3_frame f in q3mdl.frames)
	{
		mdlfile.write((char*)&type, 4);
		mdlfile.write((char*)&mincomp, 4);
		mdlfile.write((char*)&maxcomp, 4);
		mdlfile.write((char*)f.name, 16);
		
		for (surfiter = q3mdl.surfaces.begin(); surfiter != q3mdl.surfaces.end(); surfiter++)
		{
			for each (q3_vertex v in (*surfiter).frame.at(i).vertex)
			{
				coords[0] = (unsigned char) ((v.x * MD3_XYZ_SCALE - mins[0]) / (head.scale[0]));
				coords[1] = (unsigned char) ((v.y * MD3_XYZ_SCALE - mins[1]) / (head.scale[1]));
				coords[2] = (unsigned char) ((v.z * MD3_XYZ_SCALE - mins[2]) / ( head.scale[2]));
				coords[3] = findnormal(v);
				mdlfile.write((char*)&coords, 4);
//				cout << (int)coords[0] << " " << (int)coords[1] << " " << (int)coords[2] << endl;

			}
		}
		i++;
	}
}


int32_t main(int32_t argc, const char* argv[])
{
	if(argc < 2)
	{
		std::cout << "Please specify a text file" << std::endl;
		return 1;
	}

	std::ifstream cfile(argv[1]);
	if(!cfile)
	{
		std::cout << "file not found/error reading file" << std::endl;
		return 2;
	}

	std::string line;
	std::string md3path;
	std::string skinpath;
	std::string mdlpath;
	bool clampuvs;
	int32_t linecount = 0;
	while(std::getline(cfile,line))
	{
		++linecount;
		std::istringstream linestream(line);
		std::string command;
		linestream >> command;
		if (command == "md3file")
		{
			if(md3path.empty())
				linestream >> md3path;
			else std::cout << "Can only import 1 model" << std::endl;
		}
		else if (command == "skinfile")
		{
			if(skinpath.empty())
				linestream >> skinpath;
			else
				std::cout << "Can only apply 1 skin" << std::endl;
		}
		else if (command == "mdlfile")
		{
			if(mdlpath.empty())
				linestream >> mdlpath;
			else
				std::cout << "Ignoring multiple export filenames" << std::endl;
		}
		else if (command == "clampuvs")
		{
			clampuvs = true;
		}
		else if(!command.empty()) //we'll ignore entirely empty lines
			std::cout << "Bad token \"" << command << "\" on line " << linecount << std::endl;
	}
	cfile.close();
	
	if(mdlpath.empty())
		mdlpath = "output.mdl";

	std::cout << "input: \"" << md3path << "\"" << std::endl;
	std::cout << "skin:  \"" << skinpath << "\"" << std::endl;
	std::cout << "output:  \"" << mdlpath << "\"" << std::endl;

	std::ifstream q3input;
	q3input.open(md3path.c_str(), std::ios_base::binary);
	if(q3input.bad())
	{
		std::cout << "Could not open md3 file" << std::endl;
		return 2;
	}



	//load this file
	q3_model q3format;
	if(!parse_q3model(q3input, q3format))
		return 0;
	q3input.close();

	std::ifstream pcxinput;
	skin q1skin;
	pcxinput.open(skinpath.c_str(), std::ios_base::binary);
	if(pcxinput.good())
	{
		loadskin(pcxinput,q1skin);
	}//else?
	pcxinput.close();

		


	std::ofstream q1output;
	q1output.open(mdlpath.c_str(), std::ios_base::binary);

	if(!q1output.good())
	{
		std::cout << "File error, conversion terminated";
		return 0;
	}
	
	export_to_mdl(q1output, q3format, q1skin);

	q1output.close();
	
	return 0;
}
