
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm>
#include <cmath>
#include <boost/cstdint.hpp>

#define M_PI 3.14159265f


using boost::int8_t;
using boost::uint8_t;
using boost::int16_t;
using boost::int32_t;

struct pcx_header 
{
	int8_t manufacturer;
	int8_t version;
	int8_t encoding;
	int8_t bpp;
	int16_t xmin;
	int16_t ymin;
	int16_t xmax;
	int16_t ymax;
	int16_t hdpi;
	int16_t vdpi;
	int8_t colourmap[48];
	int8_t reserved;
	int8_t nplanes;
	int16_t bytesperline;
	int16_t paletteinfo;
	int16_t hscreensize;
	int16_t vscreensize;
	int8_t filler[54];
};

typedef float vec3_t [3];

vec3_t anorms_table[162] = {
#include "anorms.h"
};

struct skin
{
	int32_t width;
	int32_t height;
	char* skindata;
	skin();

	~skin();
};

skin::skin() :height(0), width(0) {};

skin::~skin()
{
	delete[] skindata;
}



class q3_header 
{
public:
	int32_t flags;
	int32_t framecount;
	int32_t surfacecount;
	int32_t ofs_frames;
	int32_t ofs_surfaces;
};

class q3_vertex 
{
public:
	int16_t x;
	int16_t y;
	int16_t z;
	int16_t norm;
};


class q3_triangle 
{
public:
	int32_t index[3];
};

class q3_frame 
{
public:
	float mins[3];
	float maxs[3];
	float org[3];
	float radius;
	int8_t name[16];
};

class q3_frame_vertices
{
public:
	std::vector <q3_vertex> vertex;
};



class q3_texcoord 
{
public:
	float u;
	float v;
};


class q3_surfaceheader
{
public:
	int32_t ident;
	int8_t name[64];
	int32_t flags;

	int32_t num_frames;
	int32_t num_shaders;
	int32_t num_verts;
	int32_t num_triangles;

	int32_t ofs_triangles;
	int32_t ofs_shaders;
	int32_t ofs_st;
	int32_t ofs_vertices;
	int32_t ofs_end;
};

class q3_surface 
{
public:
	q3_surfaceheader header;
	std::vector <q3_triangle> triangle;
	std::vector <q3_texcoord> texcoord;
	std::vector <q3_frame_vertices> frame; //note an extra layer of object here
};


class q3_model
{
public:
	q3_header header;
	std::vector <q3_frame> frames;
	std::vector <q3_surface> surfaces;
};

struct q1_header 
{
	int32_t ident;            /* magic number: "IDPO" */
	int32_t version;          /* version: 6 */

	float scale[3];         /* scale factor */
	float translate[3];     /* translation vector */
	float boundingradius;
	float eyeposition[3];   /* eyes' position */

	int32_t num_skins;        /* number of textures */
	int32_t skinwidth;        /* texture width */
	int32_t skinheight;       /* texture height */

	int32_t num_verts;        /* number of vertices */
	int32_t num_tris;         /* number of triangles */
	int32_t num_frames;       /* number of frames */

	int32_t synctype;         /* 0 = synchron, 1 = random */
	int32_t flags;            /* state flag */
	float size;
};

static float MD3_XYZ_SCALE = 1.0f / 64;